/* $LAAS: speettMainCodels.c 2008/05/27 00:28:35 tho $ */

/*
 * Copyright (c) 2008 IS/AIST-ST2I/CNRS
 *	Joint Japanese-French Robotics Laboratory (JRL).
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Sun Apr 13 2008
 */
static const char _rcsid[] = "$LAAS: speettMainCodels.c 2008/05/27 00:28:35 tho $";

#include <portLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <sys/stat.h>

#include "server/speettHeader.h"

#include "sphinx2/fbs.h"
#include "sphinx2/ad.h"
#include "sphinx2/cont_ad.h"
#include "sphinx2/dict.h"

/* this one is not exported */
extern search_hyp_t *	search_get_hyp(void);


/* --- local data ---------------------------------------------------- */

/** Default acoustic model from sphinx2 */
#define SPHINX2_HMM	SPHINX_PREFIX "/share/sphinx2/model/hmm/6k"

/** Default recording sample rate */
#define SPHINX2_SPS	16000

/** Recording handle */
static ad_rec_t *recording;

/** Continuous speech handle */
static cont_ad_t *audio;

/** Sentence stack */
char **spt_sentences_stk;
unsigned int spt_sentences_count;

/** Current dictionary */
static dictT *spt_dictionary;

/** Argument passed to sphinx engine for initialization. Borrowed from
 * the sphinx-simple demo script. Some arguments are reserved for dynamic
 * configuration (FSG and dictionary). */
static char *spt_sphinx_args[] = {
  "speet GenoM module",
#define SPHINX2_VERBOSE	2
  "-verbose",		"2",
#define SPHINX2_FSG	4
  "-fsgfn",		NULL,
#define SPHINX2_DICT	6
  "-dictfn",		NULL,
  "-live",		"TRUE",
  "-langwt",		"15",
  "-ndictfn",		SPHINX2_HMM "/noisedict",
  "-phnfn",		SPHINX2_HMM "/phone",
  "-mapfn",		SPHINX2_HMM "/map",
  "-hmmdir",		SPHINX2_HMM,
  "-hmmdirlist",	SPHINX2_HMM,
  "-8bsen",		"TRUE",
  "-sendumpfn",		SPHINX2_HMM "/sendump",
  "-cbdir",		SPHINX2_HMM
};


/* --- speettMainEnd ------------------------------------------------- */

/** Termination codel.
 *
 * \return OK or ERROR.
 */

STATUS
speettMainEnd(void)
{
  int i;

  if (recording)		ad_close(recording);
  if (SDI_F->is_sphinx_init)    fbs_end();
  SDI_F->is_sphinx_init = 0;

  for(i=0; i<spt_sentences_count; i++)
    if (spt_sentences_stk[i]) free(spt_sentences_stk[i]);
  if (spt_sentences_stk) free(spt_sentences_stk);

  spt_sentences_stk = NULL;
  spt_sentences_count = 0;

  if (spt_dictionary) {
    dict_free(spt_dictionary);
    spt_dictionary = NULL;
  }
  return OK;
}


/* --- Init ---------------------------------------------------------- */

/** Initialization request.
 *
 * Reports:      OK
 *              S_speett_NO_SUCH_FILE
 *              S_speett_NO_AUDIO_DEVICE
 *              S_speett_NO_MEMORY
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

/* speettInitMain  -  codel EXEC of Init */

ACTIVITY_EVENT
speettInitMain(spt_init_p *initp, int *report)
{
  int argc = sizeof(spt_sphinx_args)/sizeof(spt_sphinx_args[0]);
  struct stat sb;
  char fsg[1024];
  char dict[1024];

  /* cleanup previous init */
  speettMainEnd();

  /* setup sphinx engine */

  snprintf(fsg, sizeof(fsg), "%s.fsg", initp->model_base);
  snprintf(dict, sizeof(dict), "%s.dic", initp->model_base);

  if (stat(fsg, &sb)) {
    warn(fsg);
    *report = S_speett_NO_SUCH_FILE;
    return ETHER;
  }
  if (stat(dict, &sb)) {
    warn(fsg);
    *report = S_speett_NO_SUCH_FILE;
    return ETHER;
  }

  spt_sphinx_args[SPHINX2_FSG] = fsg;
  spt_sphinx_args[SPHINX2_DICT] = dict;

  fbs_init(argc, spt_sphinx_args);
  SDI_F->is_sphinx_init = 1;
  warnx("sphinx initialized from %s", initp->model_base);


  /* setup audio */

  recording = ad_open_sps(SPHINX2_SPS);
  if (!recording) {
    warnx("cannot open audio device");
    *report = S_speett_NO_AUDIO_DEVICE;
    return ETHER;
  }
  warnx("audio device initialized");


  /* load dictionary for the 'push' request */

  spt_dictionary = dict_new();
  if (!spt_dictionary) {
    warnx("cannot load dictionary");
    *report = S_speett_NO_MEMORY;
    return ETHER;
  }
  if (dict_read(spt_dictionary, dict, NULL, NULL, 0)) {
    warnx("cannot load dictionary");
    *report = S_speett_NO_MEMORY;
    return ETHER;
  }

  return ETHER;
}


/* --- Listen -------------------------------------------------------- */

/** Listen and recognize sentences
 *
 * Reports:      OK
 *              S_speett_NO_AUDIO_DEVICE
 */

/** Internal request state */
static enum {
  SPT_UTT_WAITING,
  SPT_UTT_BUFFERING,
  SPT_UTT_DECODING
} listen_state;

/* speettListenStart  -  codel START of Listen
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
speettListenStart(spt_listen_p *listenp, int *report)
{
  /* Initialize continuous listening module */
  audio = cont_ad_init(recording, ad_read);
  if (!audio) {
    warnx("cannot initialize continuous listening");
    *report = S_speett_NO_AUDIO_DEVICE;
    return ETHER;
  }

  if (ad_start_rec(recording) < 0 || cont_ad_calib(audio) < 0) {
    warnx("cannot start recording");
    *report = S_speett_NO_AUDIO_DEVICE;
    cont_ad_close(audio);
    audio = NULL;
    return ETHER;
  }

  listen_state = SPT_UTT_WAITING;
  warnx("listening...");
  return EXEC;
}

/* speettListenMain  -  codel EXEC of Listen
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
speettListenMain(spt_listen_p *listenp, int *report)
{
  static int timestamp;
  int n, done;
  int16 raw[4096];
  search_hyp_t *hyp;
  char *sentence;

  /* internal FSM */
  switch(listen_state) {
    case SPT_UTT_WAITING:
      /* await data for next utterance */
      n = cont_ad_read(audio, raw, sizeof(raw)/sizeof(raw[0]));
      if (n < 0) {
	warnx("end of audio stream");
	*report = S_speett_NO_SOUND;
	return END;
      }
      if (n == 0) return EXEC; /* sleep a bit */

      /* new data received; start recognition of new utterance */
      if (uttproc_begin_utt(NULL) < 0) {
	warnx("failed to handle new data");
	*report = S_speett_NO_SOUND;
	return END;
      }

      if (uttproc_rawdata(raw, n, 0/*non-blocking*/) < 0) {
	warnx("failed to buffer new data");
	*report = S_speett_NO_SOUND;
	return END;
      }

      listen_state = SPT_UTT_BUFFERING;
      timestamp = audio->read_ts;
      warnx("recording...");
      /*FALLTHROUGH*/;

    case SPT_UTT_BUFFERING:
      /* Read non-silence audio data until a "long" silence, >1sec */
      for(;;) {
	n = cont_ad_read(audio, raw, sizeof(raw)/sizeof(raw[0]));
	if (n < 0) {
	  warnx("end of audio stream");
	  *report = S_speett_NO_SOUND;
	  return END;

	} else if (n > 0) {
	  /* new speech data received; note current timestamp and buffer */
	  timestamp = audio->read_ts;
	  done = uttproc_rawdata(raw, n, 0);
	  if (done < 0) {
	    warnx("failed to buffer new data");
	    *report = S_speett_NO_SOUND;
	    return END;
	  }

	} else {
	  if (audio->read_ts - timestamp > SPHINX2_SPS)
	    /* no speech data; stop if more than 1 sec elapsed */
	    break;
	  /* otherwise, sleep a bit */
	  return EXEC;
	}
      }

      /* utterance ended; flush any A/D data and stop listening */
      ad_stop_rec(recording);
      cont_ad_reset(audio);

      listen_state = SPT_UTT_DECODING;
      warnx("decoding speech data");
      /*FALLTHROUGH*/;

    case SPT_UTT_DECODING:
      /* finish decoding, obtain and print result */
      uttproc_end_utt();

      if (uttproc_result(&done, &sentence, 1/*block*/) < 0)
	warnx("failed to decode data");
      else {
	/* check that the last word is in the final state */
	for(hyp = search_get_hyp(); hyp && hyp->wid >= 0; hyp++) /*empty*/;
	if (hyp && hyp[-1].wid >= 0 &&
	    hyp[-1].fsg_state != uttproc_get_fsg_final_state()) {
	  warnx("failed with best match: %s", sentence);
	  goto resume;
	}

	/* push on stack and remove leading and trailing white space */
	char **tmp = realloc(spt_sentences_stk,
			     (spt_sentences_count+1)*sizeof(spt_sentences_stk[0]));
	if (tmp) {
	  spt_sentences_stk = tmp;

	  sentence += strspn(sentence, " \t\n");
	  spt_sentences_stk[spt_sentences_count] = strdup(sentence);
	  if (spt_sentences_stk[spt_sentences_count]) {
	    sentence = strchr(spt_sentences_stk[spt_sentences_count], '\0');
	    for(sentence--; strspn(sentence, " \t\n"); sentence--)
	      *sentence = '\0';

	    sentence = spt_sentences_stk[spt_sentences_count];
	    warnx(sentence);
	    spt_sentences_count++;

	    /* if the only word is the stop word, stop listening */
	    if (!strcasecmp(sentence, listenp->stop))
	      return END;
	  }
	}
      }

    resume:
      /* resume recording for the next utterance */
      if (ad_start_rec(recording) < 0) {
	warnx("cannot start recording");
	*report = S_speett_NO_AUDIO_DEVICE;
	return END;
      }

      listen_state = SPT_UTT_WAITING;
      warnx("listening...");
      return EXEC;
  }

  /* we should not end up here */
  return END;
}

/* speettListenEnd  -  codel END and INTER of Listen
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
speettListenEnd(spt_listen_p *listenp, int *report)
{
  warnx("stopping listening activity");
  ad_stop_rec(recording);
  if (audio) cont_ad_close(audio);
  audio = NULL;
  return ETHER;
}


/* --- Pop ----------------------------------------------------------- */

/** Pop a sentence from the stack
 *
 * Reports:      OK
 *              S_speett_NO_SENTENCE
 */

/* speettPopMain  -  codel EXEC of Pop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
speettPopMain(spt_pop_p *popp, spt_pop_o *popo, int *report)
{
  int i;

  if (spt_sentences_count == 0) switch(popp->block) {
    case SPT_POP_NOBLOCK:
      popo->sentence[0] = '\0';
      *report = S_speett_NO_SENTENCE;
      return ETHER;

    case SPT_POP_BLOCK:
      return EXEC;
  }

  switch(popp->date) {
    case SPT_POP_OLDEST:
      strncpy(popo->sentence, spt_sentences_stk[0], sizeof(popo->sentence)-1);
      free(spt_sentences_stk[0]);
      for(i=1; i<spt_sentences_count; i++)
	spt_sentences_stk[i-1] = spt_sentences_stk[i];
      spt_sentences_count--;
      break;

    case SPT_POP_LATEST:
      strncpy(popo->sentence, spt_sentences_stk[spt_sentences_count-1],
	      sizeof(popo->sentence)-1);
      spt_sentences_count--;
      free(spt_sentences_stk[spt_sentences_count]);
      break;
  }

  return ETHER;
}


/* --- Push ---------------------------------------------------------- */

/** Push a sentence onto the stack (emulate user input)
 *
 * Reports:      OK
 *              S_speett_SYNTAX_ERROR
 *              S_speett_NO_MEMORY
 */

/* speettPushMain  -  codel EXEC of Push
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
speettPushMain(spt_pop_o *pushp, int *report)
{
  char *sentence, *w, *s;
  char **tmp;

  /* check syntax */
  sentence = strdup(pushp->sentence);
  if (!sentence) { *report = S_speett_NO_MEMORY; return ETHER; }

  for(w = strtok_r(sentence, " \t\n", &s); w; w = strtok_r(NULL, " \t\n", &s))
    if (dict_to_id(spt_dictionary, w) == NO_WORD) {
      warnx("unrecognized word `%s'", w);
      free(sentence);
      *report = S_speett_SYNTAX_ERROR;
      return ETHER;
    }

  /* restore origninal sentence */
  strcpy(sentence, pushp->sentence);

  /* push onto the stack */
  tmp = realloc(spt_sentences_stk,
		(spt_sentences_count+1)*sizeof(spt_sentences_stk[0]));
  if (!tmp) {
    free(sentence);
    *report = S_speett_NO_MEMORY;
    return ETHER;
  }

  spt_sentences_stk = tmp;
  spt_sentences_stk[spt_sentences_count] = sentence;
  warnx(spt_sentences_stk[spt_sentences_count]);
  spt_sentences_count++;

  return ETHER;
}
